from tqdm import tqdm
import numpy as np
import pandas as pd
from skbio import Sequence
from sklearn.base import BaseEstimator, TransformerMixin

tqdm.pandas()


class DropColumns(BaseEstimator, TransformerMixin):
    """
    Transformer to drop a group of given columns in a list
    """
    def __init__(self, columns_drop: list) -> None:
        """
        :param columns_drop: list with column names to drop
        """
        self.columns_drop = columns_drop

    def fit(self, X: pd.DataFrame, y=None) -> 'DropColumns':
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        return X.drop(self.columns_drop, axis=1)


class PickColumns(BaseEstimator, TransformerMixin):
    """
    Transformer to pick a group of given columns
    """
    def __init__(self, columns_to_pick: tuple) -> None:
        """
        :param columns_to_pick: Tuple of columns to pick
        """
        self.columns_to_pick = columns_to_pick

    def fit(self, X: pd.DataFrame, y=None) -> 'PickColumns':
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        return X[self.columns_to_pick]


class KmerVectorizer(BaseEstimator, TransformerMixin):
    """
    Sci-kit learn transformer to generate kmer column vectors using one of `sklearn.feature_extraction.text`
      vectorizers (i.e: CountVectorizer or TfidfVectorizer)
    """
    def __init__(self, vectorizer, n_kmers: int, kmer_string_precalculated_name: str = None,
                 use_progress_map: bool = False) -> None:
        self.vectorizer = vectorizer
        self.n_kmers = n_kmers
        self.use_progress_map = use_progress_map
        self.kmer_string_precalculated_name = kmer_string_precalculated_name

    def fit(self, X: pd.DataFrame, y=None) -> 'KmerVectorizer':
        # Get the kmer list into a single string separated by spaces
        if self.kmer_string_precalculated_name is None:
            if self.use_progress_map:
                kmer_sequence = X.sequences.progress_map(
                    lambda dna_seq: self.get_kmer_string(dna_seq, n_kmers=self.n_kmers))
            else:
                kmer_sequence = X.sequences.map(
                    lambda dna_seq: self.get_kmer_string(dna_seq, n_kmers=self.n_kmers))
        else:
            kmer_sequence = X[self.kmer_string_precalculated_name]

        # Fit the vectorize to the the list of kmer sequences seen in X
        self.vectorizer.fit(kmer_sequence)

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        if self.kmer_string_precalculated_name is None:
            if self.use_progress_map:
                kmer_sequence = X.sequences.progress_map(
                    lambda dna_seq: self.get_kmer_string(dna_seq, n_kmers=self.n_kmers))
            else:
                kmer_sequence = X.sequences.map(
                    lambda dna_seq: self.get_kmer_string(dna_seq, n_kmers=self.n_kmers))
        else:
            kmer_sequence = X[self.kmer_string_precalculated_name]

        X_vectorized = self.vectorizer.transform(kmer_sequence)

        kmer_names = self.vectorizer.get_feature_names()

        X_vectorized_df = pd.DataFrame(X_vectorized.toarray(),
                                       columns=kmer_names,
                                       index=X.index)

        X = X.join(X_vectorized_df)

        return X

    @staticmethod
    def get_kmer_string(sequence, n_kmers, sep=' ', overlap=True):
        """
        Function to generate strings composed of the different kmers in the sequence with separator `sep`

        :param sequence: DNA sequence string
        :param n_kmers: kmers order to look for in the sequence
        :param sep: String to use to separate the kmers found. Usually ' ' works well with vectorizers
        :param overlap: Whether to use overlap when generating the kmers r not
        :return:
        """
        kmer_sequence = sep.join(
            [str(kmer) for kmer in Sequence(sequence=sequence).iter_kmers(n_kmers, overlap=overlap)]
        )

        return kmer_sequence


class CorrelationThreshold(BaseEstimator, TransformerMixin):
    """
    Sci-kit learn transformer to generate kmer column vectors using one of `sklearn.feature_extraction.text`
      vectorizers (i.e: CountVectorizer or TfidfVectorizer)
    """
    def __init__(self, correlation_threshold: float = 0.95) -> None:
        self.correlation_threshold = correlation_threshold

    def fit(self, X: pd.DataFrame, y=None) -> 'CorrelationThreshold':
        # Get the list of heavily correlated variables to drop
        self.cols_to_drop = self.cols_drop(X_train=X)
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X = X.drop(X.columns[self.cols_to_drop], axis=1)
        return X

    def cols_drop(self, X_train: pd.DataFrame) -> list:
        """
        Picks columns to drop (one out of those with a correlation higher than `self.correlation_threshold`)

        :param X_train: Dataframe with features
        :return:
        """
        cor_matrix = X_train.corr().abs()
        upper_tri = cor_matrix.where(np.triu(np.ones(cor_matrix.shape), k=1).astype(np.bool))
        cols_to_drop = [column for column in upper_tri.columns if any(upper_tri[column] > self.correlation_threshold)]

        return cols_to_drop


