import os

from typing import Tuple

import pandas as pd
from sklearn.metrics import roc_curve, auc
import matplotlib.pyplot as plt
import scikitplot as skplt


def performance_plots(estimator, X: pd.DataFrame, y_true: pd.Series, object_dir: str, prefix: str = 'train',
                      fig_size: Tuple[int, int] = (15, 8)) -> None:
    """
    Save plots used to verify perfroance of the classifier such as recall-precision curves, ROC, and normalized and
    non-normalized confusion matrix

    :param estimator: model for which we want to obtain the plots
    :param X: Features on which we want to generate the predictions (train or test)
    :param y_true: Ground truth
    :param object_dir: path of the root directory to create a dir where we will save the plots
    :param prefix: prefix to add to the directory's name where we will add the plots
    :param fig_size: 2D size of the plots that will be saved
    :return:
    """
    y_probas = estimator.predict_proba(X)
    y_pred = estimator.predict(X)

    plot_parmas_dict = dict(figsize=fig_size, text_fontsize=15, title_fontsize=20)
    # Precision-Recall curve
    skplt.metrics.plot_precision_recall_curve(y_true, y_probas,
                                              title=f'{prefix}: Precision-Recall Curve',
                                              **plot_parmas_dict)
    plt.savefig(os.path.join(object_dir, f'{prefix}_precision_recall_curve.png'))

    # Normalized confusion matrix
    skplt.metrics.plot_confusion_matrix(y_true, y_pred, normalize=True,
                                        title=f'{prefix}: Normalized Confusion Matrix',
                                        **plot_parmas_dict)
    plt.savefig(os.path.join(object_dir, f'{prefix}_normalized_confusion_matrix.png'))

    # Count confusion matrix
    skplt.metrics.plot_confusion_matrix(y_true, y_pred, normalize=False,
                                        title=f'{prefix}: Count Confusion Matrix',
                                        **plot_parmas_dict)
    plt.savefig(os.path.join(object_dir, f'{prefix}_count_confusion_matrix.png'))

    # Plot ROC
    fpr, tpr, _ = roc_curve(y_true.ravel(), y_probas[:, 1].ravel())
    roc_auc = auc(fpr, tpr)

    plt.figure(figsize=fig_size)
    lw = 2
    plt.plot(fpr, tpr, color='darkorange',
             lw=lw, label='ROC curve (area = %0.2f)' % roc_auc)
    plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate', fontsize=plot_parmas_dict['text_fontsize'])
    plt.ylabel('True Positive Rate', fontsize=plot_parmas_dict['text_fontsize'])
    plt.title(f'{prefix}: Receiver operating characteristic example', fontsize=plot_parmas_dict['title_fontsize'])
    plt.legend(loc="lower right")
    plt.savefig(os.path.join(object_dir, f'{prefix}_auc_roc_curve.png'))
