import os
import json

from typing import Tuple, Optional

import numpy as np
import pandas as pd
from sklearn.metrics import roc_curve, auc, average_precision_score, roc_auc_score
import matplotlib.pyplot as plt
import scikitplot as skplt


def performance_plots(estimator: Optional,
                      X: Optional[pd.DataFrame], y_true: pd.Series,
                      object_dir: Optional[str], prefix: str = 'train',
                      y_pred_prob: Optional[np.array] = None, threshold: Optional[float] = None,
                      fig_size: Tuple[int, int] = (15, 8)) -> None:
    """
    Save plots used to verify perfroance of the classifier such as recall-precision curves, ROC, and normalized and
    non-normalized confusion matrix

    :param estimator: model for which we want to obtain the plots
    :param X: Features on which we want to generate the predictions (train or test)
    :param y_true: Ground truth
    :param y_pred_prob: Predictions if already available
    :param threshold: Threshold to obtain classes for y_pred_prob and use to calculate the confusion matrix
    :param object_dir: path of the root directory to create a dir where we will save the plots
    :param prefix: prefix to add to the directory's name where we will add the plots
    :param fig_size: 2D size of the plots that will be saved
    :return:
    """
    if estimator is not None:
        y_probas = estimator.predict_proba(X)
        y_pred = estimator.predict(X)
    else:
        y_pred = (y_pred_prob > threshold).astype(int)
        y_probas = np.concatenate([1 - y_pred_prob, y_pred_prob], axis=1)

    # Print the predictions of the model and store them in a json
    auc_prc = average_precision_score(y_true=y_true, y_score=y_pred_prob)
    print(f'auc_prc: {auc_prc}')
    auc_roc = roc_auc_score(y_true=y_true, y_score=y_pred_prob)
    print(f'auc_roc: {auc_roc}')

    if object_dir is not None:
        json.dump({'auc_roc': auc_roc, 'auc_prc': auc_prc},
                  open(os.path.join(object_dir, f'metrics.json'), 'wb'))

    plot_parmas_dict = dict(figsize=fig_size, text_fontsize=15, title_fontsize=20)
    # Precision-Recall curve
    skplt.metrics.plot_precision_recall_curve(y_true, y_probas,
                                              title=f'{prefix}: Precision-Recall Curve',
                                              **plot_parmas_dict)
    if object_dir is not None:
        plt.savefig(os.path.join(object_dir, f'{prefix}_precision_recall_curve.png'))

    # Normalized confusion matrix
    skplt.metrics.plot_confusion_matrix(y_true, y_pred, normalize=True,
                                        title=f'{prefix}: Normalized Confusion Matrix',
                                        **plot_parmas_dict)
    if object_dir is not None:
        plt.savefig(os.path.join(object_dir, f'{prefix}_normalized_confusion_matrix.png'))

    # Count confusion matrix
    skplt.metrics.plot_confusion_matrix(y_true, y_pred, normalize=False,
                                        title=f'{prefix}: Count Confusion Matrix',
                                        **plot_parmas_dict)
    if object_dir is not None:
        plt.savefig(os.path.join(object_dir, f'{prefix}_count_confusion_matrix.png'))

    # Plot ROC
    fpr, tpr, _ = roc_curve(y_true.ravel(), y_probas[:, 1].ravel())
    roc_auc = auc(fpr, tpr)

    plt.figure(figsize=fig_size)
    lw = 2
    plt.plot(fpr, tpr, color='darkorange',
             lw=lw, label='ROC curve (area = %0.2f)' % roc_auc)
    plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate', fontsize=plot_parmas_dict['text_fontsize'])
    plt.ylabel('True Positive Rate', fontsize=plot_parmas_dict['text_fontsize'])
    plt.title(f'{prefix}: Receiver operating characteristic example', fontsize=plot_parmas_dict['title_fontsize'])
    plt.legend(loc="lower right")

    if object_dir is not None:
        plt.savefig(os.path.join(object_dir, f'{prefix}_auc_roc_curve.png'))
